'''
Object to handle and deal with prior info for retrieval
'''

import numpy as np
from collections.abc import Iterable

from ._params import _Param, _UniformParam, _GaussianParam
from .detrending_funcs import NthOrderDetrendingFunction
from .detrender import DetrendingFunction
from ._limb_darkening_handler import LimbDarkeningHandler
from ._paramarray import ParamArray

# Parameters which do not vary with telescope, filter, or epoch
# Note that t0 is in here, as it is global unless ttv mode is on
_global_params = ['P', 'a', 'inc', 'ecc', 'w', 't0']

# Parameters which vary with filter
_filter_dependent_params = ['rp', 'q0', 'q1', 'q2', 'q3']

# Params which vary with epoch
_epoch_dependent_params = []

# The possible ldc lables
_all_ldcs = ['q0', 'q1', 'q2', 'q3']

# Default values for parameters (used in io.parse_priors_list)
_prior_info_defaults = {'P':1, 'a':10, 'inc':90, 'rp':0.05, 't0':0, 'ecc':0,
                        'w':90, 'limb_dark':'quadratic', 'q0':0.1, 'q1':0.3,
                        'q2':None, 'q3':None, 'n_telescopes':1,  'n_filters':1,
                        'n_epochs':1, 'norm':1}

def setup_priors(P, t0, a, rp, inc, ecc, w, limb_dark, n_telescopes, n_filters,
                 n_epochs, q0=None, q1=None, q2=None, q3=None, allow_ttv=False,
                 lightcurves=None):
    '''
    Factory function to initialise a PriorInfo object

    qX should either be a single value or a list of values length n_filters
    '''
    # Deal with the qX values
    if q0 is not None and not isinstance(q0, Iterable):
        q0 = [q0 for i in range(n_filters)]
    if q1 is not None and not isinstance(q1, Iterable):
        q1 = [q1 for i in range(n_filters)]
    if q2 is not None and not isinstance(q2, Iterable):
        q2 = [q2 for i in range(n_filters)]
    if q3 is not None and not isinstance(q3, Iterable):
        q3 = [q3 for i in range(n_filters)]

    default_dict = {'P':P, 't0':t0, 'a':a, 'rp':rp, 'inc':inc, 'ecc':ecc,
                    'w':w, 'q0':q0, 'q1':q1, 'q2':q2, 'q3':q3, 'norm':1}

    return PriorInfo(default_dict, limb_dark, n_telescopes, n_filters,
                      n_epochs, allow_ttv, lightcurves)



class PriorInfo:
    '''
    Object to deal with anything involving priors

    Parameters
    ----------
    default_dict : dictionary
        Dictionary containing default parameter values.
    limb_dark : str
        The limb darkening model to use.
    n_telescopes : int
        The number of different telescopes used in the data set
    n_filters : int
        The number of different filters used in the data set
    n_epochs : int
        The number of different epochs used in the data set
    allow_ttv : bool, optional
        If True, will fit t0 to each transit. Default is False.
    lightcurves : array_like, shape(n_telescopes, n_filters, n_epochs)
        Array of the light curves. If there is no observation for a particular
        telescope, filter, epoch combination, then the value should be set to
        None.
    '''
    def __init__(self, default_dict, limb_dark, n_telescopes, n_filters,
                 n_epochs, allow_ttv=False, lightcurves=None):
        # Store the basics
        self.limb_dark = limb_dark
        self.n_telescopes = n_telescopes
        self.n_filters = n_filters
        self.n_epochs = n_epochs
        self.allow_ttv = allow_ttv

        # Initialise limb darkening (set to off)
        self.ld_handler = LimbDarkeningHandler(self.limb_dark)
        self.limb_dark_coeffs = self.ld_handler.get_required_coefficients()
        self.fit_ld = False
        self.ld_fit_method = 'off'

        # Initialse detrending (set to off)
        self.detrend = False
        self.detrending_coeffs = []

        # Initialse normalisation (set to off)
        self.normalise=False

        #####################
        # Set up the priors #
        #####################
        # Dictionary containing info on all the priors
        self.priors = {}

        for key in default_dict.keys():
            #print('Param:', key)
            # Loop through each parameter and initialise the prior


            if key in _global_params:
                # These are global parameters
                if key == 't0' and allow_ttv:
                    # We are fitting a separate t0 for each epoch
                    self.priors[key] = ParamArray(key, (1,1, self.n_epochs), False, False, True, default_dict[key])
                else:
                    self.priors[key] = ParamArray(key, (1,1, 1), False, False, False, default_dict[key])

            elif key in _filter_dependent_params:
                # These parameters vary with filter
                shape = (1, self.n_filters, 1)

                # Deal with the ldcs
                if key in _all_ldcs:
                    if key in self.limb_dark_coeffs:
                        #print('Initialising ldc', key)
                        self.priors[key] = ParamArray(key, shape, False, True, False, default_dict[key][0])
                        for qi, q in enumerate(default_dict[key]):
                            self.priors[key].set_value(q, filter_idx=qi)

                # Otherwise, this is rp
                else:
                    self.priors[key] = ParamArray(key, shape, False, True, False, default_dict[key])

            else:
                # This is a lightcurve dependent parameter
                self.priors[key] = ParamArray(key, (self.n_telescopes, self.n_filters,self.n_epochs), True, True, True, default_dict[key], lightcurves=lightcurves)

        # All priors should now be initialised

        # We also need a list for keeping track of what is being fitted
        # Each entry in this will be
        # [param name, telescope index, filter index, epoch index]
        self.fitting_params = None

        #for p in self.priors.keys():
        #    print(p, self.priors[p].array)

    ###############################################################
    #                   ADDING FIT PARAMS                         #
    ###############################################################

    def add_uniform_fit_param(self, name, low_lim, high_lim,
                              telescope_idx=None, filter_idx=None, epoch_idx=None):
        '''
        Adds a new parameter which will be fitted uniformly in the range given
        by low_lim and high_lim
        '''
        if name in ['a', 'P', 'rp', 'inc', 'ecc', 'w']:
            negative_allowed = False
        else:
            negative_allowed = True

        self.priors[name].add_uniform_fit_param(low_lim, high_lim,
                                                telescope_idx, filter_idx,
                                                epoch_idx, negative_allowed)

        # Store some info for later
        if self.fitting_params is None:
            self.fitting_params = np.array([[name, telescope_idx, filter_idx, epoch_idx]], dtype=object)
        else:
            self.fitting_params = np.append(self.fitting_params, np.array([[name, telescope_idx, filter_idx, epoch_idx]], object), axis=0)
        #self.fitting_params.append([name, telescope_idx, filter_idx, epoch_idx])

    def add_gaussian_fit_param(self, name, mean, stdev,
                               telescope_idx=None, filter_idx=None, epoch_idx=None):
        '''
        Adds a new parameter which will be fitted with a Gaussian prior
        '''
        # If this is t0 and ttvs are on, we want to make the prior much wider to allow for proper TTVs
        #if name == 't0' and self.allow_ttv:
            # Set the Gaussian width to 0.1 days for ttv fitting mode
            #stdev = 0.1
        if name in ['a', 'P', 'rp', 'inc', 'ecc', 'w']:
            negative_allowed = False
        else:
            negative_allowed = True

        self.priors[name].add_gaussian_fit_param(mean, stdev,
                                                 telescope_idx, filter_idx,
                                                 epoch_idx, negative_allowed)

        # Store some info for later
        if self.fitting_params is None:
            self.fitting_params = np.array([[name, telescope_idx, filter_idx, epoch_idx]], dtype=object)
        else:
            self.fitting_params = np.append(self.fitting_params, np.array([[name, telescope_idx, filter_idx, epoch_idx]], object), axis=0)


    ###############################################################
    #         ADDING DETRENDING/NORMALISATION/LDC FITTING         #
    ###############################################################

    def fit_detrending(self, lightcurves, method_list, method_index_array,
                       limits=None):
        '''
        Intialises detrending
        '''
        if self.detrend:
            raise ValueError('Detrending is already initialised. You need to make a new PriorInfo to use another detrending method!')

        # Store some info - used in splitting
        self._detrend_method_list = method_list
        self._detrend_method_index_array = method_index_array

        self.detrending_coeffs = [[] for method in method_list]

        # Go through each lightcurve and initialise the relevant detrending
        for i in np.ndindex(lightcurves.shape):
            telescope_idx, filter_idx, epoch_idx = i

            if lightcurves[i] is not None:
                # We have a light curve, pull out the method info
                method_idx = method_index_array[i]
                method = method_list[method_idx]
                low_lim = limits[method_idx][0]
                high_lim = limits[method_idx][1]

                if method[0] == 'off':
                    # No detrending - skip this curve
                    pass
                else:
                    # First we set up the detrending for the lightcurve
                    if method[0] == 'nth order':
                        lightcurves[i].set_detrending(method[0], order=method[1], method_idx=method_idx)
                    elif method[0] == 'custom':
                        lightcurves[i].set_detrending(method[0], function=method[1], method_idx=method_idx)

                    # Now set up the required parameters to be fitted
                    n_coeffs = lightcurves[i].n_detrending_params

                    for coeff_i in range(n_coeffs):
                        coeff_name = 'd{}_{}'.format(coeff_i, method_idx)

                        if coeff_name not in self.priors:
                            # Need to initialise the entry in the priors dict
                            self.detrending_coeffs[method_idx].append(coeff_name)

                            # Have to deal with nth order and custom separately
                            if method[0] == 'nth order':
                                shape = (self.n_telescopes, self.n_filters, self.n_epochs)

                                # Initialise the prior
                                self.priors[coeff_name] = ParamArray(coeff_name, shape, True, True, True)

                            elif method[0] == 'custom':
                                # Work out telescope, filter, epoch dependencies
                                tel_dep = coeff_i + 1 in method[2]
                                filt_dep = coeff_i + 1 in method[3]
                                epoch_dep = coeff_i + 1 in method[4]

                                # Work out the shape needed for the ParamArray
                                shape = ([1, self.n_telescopes][tel_dep],
                                         [1, self.n_filters][filt_dep],
                                         [1, self.n_epochs][epoch_dep])

                                # Initialise the prior
                                self.priors[coeff_name] = ParamArray(coeff_name, shape, tel_dep, filt_dep, epoch_dep)


                        # Now set up the fitting. We assume a uniform prior
                        self.add_uniform_fit_param(coeff_name, low_lim, high_lim, telescope_idx, filter_idx, epoch_idx)

        self.detrend=True

    def fit_limb_darkening(self, fit_method='independent',
                           host_T=None, host_logg=None,
                           host_z=None, filters=None,
                           n_samples=20000, do_mc=False, cache_path=None):
        '''
        Initialises fitting of limb darkening parameters, either independently
        or coupled across wavebands.

        Parameters
        ----------
        fit_method : str, optional
            Determines the mode of fitting of LD parameters. The available
            modes are:
                - 'coupled' : all limb darkening parameters are fitted
                  independently, but are coupled to a wavelength dependent
                  model based on the host parameters
                - 'single' : LD parameters are still tied to a model, but only
                  the first filter is actively fitted. The remaining filters
                  are estimated based off the ratios given by ldtk for a host
                  with the given parameters. This mode is useful for a large
                  number of filters, as 'coupled' or 'independent' fitting will
                  lead to much higher computation times.
                - 'independent' : Each LD coefficient is fitted separately for
                  each filter, with no coupling to the ldtk models.
            The default is 'independent'.
        host_T : tuple or None, optional
            The effective temperature of the host star, in Kelvin, given as a
            (value, uncertainty) pair. Must be provided if fit_method is
            'coupled' or 'single'. Default is None.
        host_logg : tuple or None, optionalor None, optional
            The log_10 of the surface gravity of the host star, with gravity
            measured in cm/s2. Should be given as a (value, uncertainty) pair.
            Must be provided if fit_method is 'coupled' or 'single'. Default is
            None.
        host_z : tuple or None, optional
            The metalicity of the host, given as a (value, uncertainty) pair.
            Must be provided if fit_method is 'coupled' or 'single'. Default is
            None.
        filters : array_like or None, optional
            The set of filters, given in [low, high] limits for the wavelengths
            with the wavelengths given in nanometers. The ordering of the
            filters should correspond to the filter_idx parameter used
            elsewhere. Must be provided if fit_method is 'coupled' or 'single'.
            Default is None.
        n_samples : int, optional
            The number of limb darkening profiles to create. Passed to
            ldtk.LDPSetCreator.create_profiles(). Default is 20000.
        do_mc : bool, optional
            If True, will use MCMC to estimate coefficient uncertainties more
            accurately. Default is False.
        cache_path : str, optional
            This is the path to cache LDTK files to. If not specified, will
            default to the LDTK default
        '''
        # Sanity checks
        if fit_method not in ['coupled', 'single', 'independent']:
            raise ValueError('Unrecognised fit method {}'.format(fit_method))

        if not fit_method == 'independent':
            if filters is None:
                raise ValueError('Filters must be provided for coupled and single ld_fit_methods')
            if not len(filters) == self.n_filters:
                raise ValueError('{} filters were given, but there are {} filters required!'.format(len(filters), self.n_filters))

        # Store some useful information
        self.fit_ld = True
        self.ld_fit_method = fit_method
        self.filters = filters
        self._n_ld_samples = n_samples
        self._do_ld_mc = do_mc
        self._ld_cache_path = cache_path

        # Set up fitting for each LDC
        for i in range(self.n_filters):
            for ldc, name in enumerate(self.limb_dark_coeffs):
                self.add_uniform_fit_param(name, 0, 1, filter_idx=i)

            # If we are in 'single' mode, we only need to fit for the first
            # wavelength
            if self.ld_fit_method == 'single':
                break

        if not fit_method == 'independent':
            # Now if we are coupling across wavelength we must initialise PyLDTK
            self.ld_handler.initialise_ldtk(host_T, host_logg, host_z, filters,
                                            self.limb_dark,
                                            n_samples, do_mc, cache_path)

    def fit_normalisation(self, lightcurves):
        '''
        When run, the Retriever will fit normalisation of the data as a
        free parameter.

        Parameters
        ----------
        lightcurves : np.array of `LightCurve`s, shape (n_telescopes, n_filters, n_epochs)
            The array of LightCurves to be normalised. We can use the value of
            fluxes to estimate a normalisation constant (c_n) range for each
            light curve. We use
                ``1/f_median -1 <= c_n <= 1/f_median + 1``
            as the default range, where f_median is the median flux value.
        '''
        if self.normalise:
            raise ValueError('Detrending is already initialised. You need to make a new PriorInfo to use another detrending method!')

        for i in np.ndindex(lightcurves.shape):
            telescope_idx, filter_idx, epoch_idx = i

            if lightcurves[i] is not None:
                # A light curve exists. Set up normalisation
                best, low, high = lightcurves[i].set_normalisation()
                self.add_uniform_fit_param('norm', low, high, telescope_idx, filter_idx, epoch_idx)

        self.normalise = True

    ###############################################################
    #             CONVERSIONS FOR FITTING ROUTINES                #
    ###############################################################

    def _convert_unit_cube(self, cube):
        '''
        Takes the unit cube provided by dynesty (all values between 0 and 1)
        and converts them into physical values
        '''
        new_cube = np.zeros(len(self.fitting_params))

        for i, param_info in enumerate(self.fitting_params):
            name, tidx, fidx, eidx = param_info
            new_cube[i] = self.priors[name].from_unit_interval(cube[i], tidx, fidx, eidx)

        return new_cube

    def _interpret_param_array(self, array):
        '''
        Interprets the parameter cube generated by dynesty and returns them in
        a format usable by the LikelihoodCalculator
        '''

        # Initialse a dictionary to store the physical results in
        result = {}

        for name in self.priors.keys():
            # First we initialise the entry in the results dictionary
            result[name] = self.priors[name].generate_blank_ParamArray()

        initialised_params = []
        # Now we add in the results
        for i, param_info in enumerate(self.fitting_params):
            name, tidx, fidx, eidx = param_info
            result[name][tidx, fidx, eidx] = array[i]
            initialised_params.append(name)

        # If things aren't being fitted, add these in too
        for name in result.keys():
            if name not in initialised_params:
                for i in np.ndindex(self.priors[name].shape):
                    result[name][i] = self.priors[name][i]

        return result

    def _interpret_final_results(self, results):
        '''
        Generates a dictionary of results and a dictionary of errors from the
        final results of a run
        '''
        best_results = results.best
        best_result_errors = results.uncertainties

        result_dict = {}
        errors_dict = {}

        # Initialise the entries in the dictionaries
        for name in self.priors.keys():
            # First we initialise the entry in the results dictionary
            result_dict[name] = self.priors[name].generate_blank_ParamArray()
            errors_dict[name] = self.priors[name].generate_blank_ParamArray()

        initialised_params = []
        # Go through the results object and add in the results - only includes
        # fitted parameters
        for i, param_info in enumerate(self.fitting_params):
            name, tidx, fidx, eidx = param_info
            result_dict[name][tidx, fidx, eidx] = best_results[i]
            errors_dict[name][tidx, fidx, eidx] = best_result_errors[i]
            initialised_params.append(name)

        # Now go through and fill in any non-fitted parameters
        for name in result_dict.keys():
            if name not in initialised_params:
                for i in np.ndindex(self.priors[name].shape):
                    result_dict[name][i] = self.priors[name][i]
                    errors_dict[name][i] = 0

        return result_dict, errors_dict


    ###############################################################
    #                          MISC                               #
    ###############################################################
    def __str__(self):
        print_str = 'Priors:\n'
        print_str += 'Limb darkening model: {}\n'.format(self.limb_dark)
        print_str += 'n telescopes: {}\n'.format(self.n_telescopes)
        print_str += 'n filters: {}\n'.format(self.n_filters)
        print_str += 'n epochs: {}\n'.format(self.n_epochs)
        for var in self.priors:
            print_str += self.priors[var].__str__()
        print_str += 'Total {} fitting parameters'.format(len(self.fitting_params))
        return print_str



'''
Class to handle limb darkening parameters through PyLDTK

'''

import numpy as np
from ldtk import LDPSetCreator, BoxcarFilter, TabulatedFilter
import os
from collections.abc import Iterable

_implemented_ld_models = ['linear', 'quadratic', 'nonlinear', 'power2', 'squareroot']

class LDTKHandler:
    '''
    The LDTKHandler provides an easy way to interface ldtk with TransitFit.

    Parameters
    ----------
    host_T : tuple
        The effective temperature of the host star, in Kelvin, given as a
        (value, uncertainty) pair.
    host_logg : tuple
        The log_10 of the surface gravity of the host star, with gravity
        measured in cm/s2. Should be given as a (value, uncertainty) pair.
    host_z : tuple
        The metalicity of the host, given as a (value, uncertainty) pair.
    filters : array_like
        The set of filters, given in [low, high] limits for the wavelengths
        with the wavelengths given in nanometers if a uniform filter is to
        be used, or [[wavelength...], [transmission]] if a fully-defined
        profile is being used. The ordering of the filters should
        correspond to the filter_idx parameter used elsewhere.
    ld_method : str, optional
        The model of limb darkening to use. Allowed values are 'linear',
        'quadratic', 'squareroot', 'power2', and 'nonlinear'. Default is
        'quadratic'.
    n_samples : int, optional
        The number of limb darkening profiles to create. Passed to
        ldtk.LDPSetCreator.create_profiles(). Default is 20000.
    do_mc : bool, optional
        If True, will use MCMC to estimate coefficient uncertainties more
        accurately. Default is False.
    cache_path : str, optional
        This is the path to cache LDTK files to. If not specified, will
        default to the LDTK default
    '''
    def __init__(self, host_T, host_logg, host_z, filters,
                 ld_model='quadratic', n_samples=20000, do_mc=False,
                 cache_path=None):

        # Sanity checks
        if not ld_model in _implemented_ld_models:
            raise ValueError('Unrecognised ld_model {}'.format(ld_model))

        self.default_model = ld_model

        # Set up the filters
        #print('Setting up filters')
        ldtk_filters = []
        for i, f in enumerate(filters):
            if isinstance(f[0], Iterable):
                # We have been passed a full filter profile, set up
                # TabulatedFilter
                # Work out if the profile is in percent or fraction - is
                # anything bigget than 1?
                if np.any(f[1] > 1):
                    tmf = 1e-2
                else:
                    tmf = 1
                ldtk_filters.append(TabulatedFilter(i, f[0], f[1], tmf))
            else:
                ldtk_filters.append(BoxcarFilter(i, f[0], f[1]))

        # Make the set creator, downloading data files if required
        if cache_path is not None:
            os.makedirs(cache_path, exist_ok=True)
        #print('Making LD parameter set creator.')
        #print('This may take some time as we may need to download files...')
        self.set_creator = LDPSetCreator(teff=host_T, logg=host_logg, z=host_z,
                                    filters=ldtk_filters, cache=cache_path)

        # Get the LD profiles from the set creator
        #print('Obtaining LD profiles')
        self.profile_set = self.set_creator.create_profiles(nsamples=n_samples)

        # Find the 'best values' for each filter and then find the ratios
        # compared to the first.
        #print('Finding coefficients and ratios')
        self.coeffs = {}
        self.ratios = {}

        self._power2_available = True

        for model in _implemented_ld_models:
            try:
                self.coeffs[model] = self._extract_best_coeffs(model)
                self.ratios[model] = self.coeffs[model][0] / self.coeffs[model][0][0]
            except Exception as e:
                print(e)
                print(f'Unable to initialise {model} model')
                self._power2_available = False

    def estimate_values(self, ld0_values, ld_model):
        '''
        If given a set of LD param values for filter 0, will estimate the LD
        parameters for all filters based on the ratios between the best values
        found in initialisation.

        Parameters
        ----------
        ld1_values : float or array_like
            The LD parameters for filter 0
        ld_model : str
            The limb darkening model to use

        Returns
        -------
        all_ld_values : array_like, shape (n_filters, n_coeffs)
            The estimated limb darkening parameters
        '''
        if ld_model == 'power2' and not self._power2_available:
            raise ValueError('power2 model is not available. If you want to use this, please use the development version of ldtk available on https://github.com/hpparvi/ldtk, rather than the pypi version.')

        return ld0_values * self.ratios[ld_model]

    def _extract_best_coeffs(self, ld_model):
        '''
        Extracts the best values for a limb darkening model for the filters

        Parameters
        ----------
        ld_model : str
            The limb darkening model to obtain the values for.

        Returns
        -------
        coeffs : array_like, shape (n_filters, n_coeffs)
            The coefficients for each filter
        err : array_like, shape (n_filters, n_coeffs)
            The uncertainty on each of the coefficients
        '''
        if ld_model == 'linear':
            coeff, err = self.profile_set.coeffs_ln()
        elif ld_model == 'quadratic':
            coeff, err = self.profile_set.coeffs_qd()
        elif ld_model == 'nonlinear':
            coeff, err = self.profile_set.coeffs_nl()
        elif ld_model == 'power2':
            if not self._power2_available:
                raise ValueError('power2 model is not available. If you want to use this, please use the development version of ldtk available on https://github.com/hpparvi/ldtk, rather than the pypi version.')
            coeff, err = self.profile_set.coeffs_p2()
        elif ld_model == 'squareroot':
            coeff, err = self.profile_set.coeffs_sq()

        else:
            raise ValueError('Unrecognised ld_model {}'.format(ld_model))

        return coeff, err

    def lnlike(self, coeffs, ld_model):
        '''
        Evaluates the log likelihood for a set of coefficients

        Parameters
        ----------
        coeffs : array_like, shape (n_filters, n_coeffs)
            The coefficients to evaluate the log likelihood for.
        ld_model : str, optional
            The model to use. Defaults to self.default_model

        '''
        if ld_model == 'linear':
            return self.profile_set.lnlike_ln(coeffs)
        if ld_model == 'quadratic':
            return self.profile_set.lnlike_qd(coeffs)
        if ld_model == 'nonlinear':
            return self.profile_set.lnlike_nl(coeffs)
        if ld_model == 'power2':
            if not self._power2_available:
                raise ValueError('power2 model is not available. If you want to use this, please use the development version of ldtk available on https://github.com/hpparvi/ldtk, rather than the pypi version.')
            return self.profile_set.lnlike_p2(coeffs)
        if ld_model == 'squareroot':
            return self.profile_set.lnlike_sq(coeffs)

        raise ValueError('Unrecognised ld_model {}'.format(ld_model))



'''
A function which will run everything when given a path to light curve and
priors!
'''

from .io import read_priors_file, read_input_file, read_filter_info, parse_data_path_list, read_data_path_array, parse_priors_list, parse_filter_list
from .retriever import Retriever

import numpy as np
import os




def run_retrieval(data_files, priors, filter_info=None,
                  detrending_list=[['nth order', 1]],
                  limb_darkening_model='quadratic',
                  ld_fit_method='independent', fitting_mode='auto',
                  max_batch_parameters=25, batch_overlap=2,
                  host_T=None, host_logg=None, host_z=None, host_r=None,
                  nlive=300, dlogz=None, maxiter=None, maxcall=None,
                  dynesty_sample='rslice', dynesty_bounding='multi',
                  normalise=True, detrend=True,
                  results_output_folder='./output_parameters',
                  final_lightcurve_folder='./fitted_lightcurves',
                  summary_file='summary_output.csv',
                  full_output_file='full_output.csv',
                  plot_folder='./plots', plot=True,
                  marker_color='dimgrey', line_color='black', ldtk_cache=None,
                  n_ld_samples=20000, do_ld_mc=False, data_skiprows=0,
                  allow_ttv=False, filter_delimiter=None,
                  detrending_limits=None, bin_data=False, cadence=2,
                  binned_color='red', walks=100, slices=10):
    '''
    Runs a full retrieval of posteriors using nested sampling on a transit
    light curve or a set of transit light curves. For more guidance on the use
    of input files and structuring, see the TransitFit documentation.

    Parameters
    ----------
    data_files : str
        The path to the data input .csv file, which contains the paths to the
        light curve data.

    priors : str
        Path to a .csv file containing prior information on each variable to be
        fitted.

    filter_info : str, optional
        Path to a .csv file containing information on the wavelengths of the
        filters that observations were made at.

        This is required if ld_fit_method is `'single'` or `'coupled'`. If not
        None and host_T, host_logg and host_z are not None, retrieval will
        include fitting realistic limb darkening parameters for the filters.
        Default is None.

    detrending_list : array_like, shape (n_detrending_models, 2)
        A list of different detrending models. Each entry should consist
        of a method and a second parameter dependent on the method.
        Accepted methods are
            ``['nth order', order]``

            ``['custom', function, [global fit indices, filter fit indices, epoch fit indices]]``

            ``['off', ]``
        ``function`` here is a custom detrending function. TransitFit assumes
        that the first argument to this function is times and that all
        other arguments are single-valued - TransitFit cannot fit
        list/array variables. If 'off' is used, no detrending will be
        applied to a light curve using this model.

        If a custom function is used, and some inputs to the function
        should not be fitted individually for each light curve, but should
        instead be shared either globally, within a given filter, or within
        a given epoch, the indices of where these fall within the arguments
        of the detrending function should be given as a list. If there are
        no indices to be given, then use an empty list: []
        e.g. if the detrending function is given by::

            foo(times, a, b, c):
                # do something

        and a should be fitted globally, then the entry in the method_list
        would be ``['custom', foo, [1], [], []]``.

    ld_model : str, optional
        The limb darkening model to use. Allowed models are

            - ``'linear'``
            - ``'quadratic'``
            - ``'squareroot'``
            - ``'power2'``
            - ``'nonlinear'``

        With the exception of the non-linear model, all models are constrained
        by the method in Kipping (2013), which can be found at
        https://arxiv.org/abs/1308.0009.
        Default is 'quadratic'.

    ld_fit_method : str, optional
        Determines the mode of fitting of limb darkening parameters. The
        available modes are:
            - ``'coupled'`` : all limb darkening parameters are fitted
              independently, but are coupled to a wavelength dependent
              model based on the host parameters through `ldkt`
            - ``'single'`` : LD parameters are still tied to a model, but only
              the first filter is actively fitted. The remaining filters
              are estimated based off the ratios given by ldtk for a host
              with the given parameters. This mode is useful for a large
              number of filters, as 'coupled' or 'independent' fitting will
              lead to much higher computation times.
            - ``'independent'`` : Each LD coefficient is fitted separately for
              each filter, with no coupling to the ldtk models.
            - ``'off'`` : no limb darkening fitting will occurr. If using this
              mode, it it strongly recommended to set values for the
              Kipping q parameters using the priors file.
        Default is `'independent'`

    fitting_mode : {``'auto'``, ``'all'``, ``'2_stage'``, ``'folded'``, ``'batched'``}, optional
        The approach TransitFit takes towards limiting the number of parameters
        being simultaneously fitted. The available modes are:
        - ``'auto'`` : Will calculate the number of parameters required to fit
          all the data simulataneously. If this is less than max_parameters,
          will set to ``'all'`` mode, else will set to ``'folded'`` if at least
          one filter has at least 3 epochs in it. Otherwise will set to
          ``'batched'``
        - ``'all'`` : Fits all parameters simultaneously, with no folding or
          batching of curves. Should be used with caution when fitting very
          large (~< 30) numbers of parameters.
        - ``'2_stage'`` : Fits in 2 stages, first detrending the light curves
          and then fitting the detrended curves simultaneously, using the
          ``'batched'`` approach if required.
        - ``'folded'`` : Useful for fitting curves with multiple epochs for each
          filter. TransitFit will fit each filter separately and produce a
          period-folded light curve for each filter, before fitting these
          simultaneously, using the ``'batched'`` approach if required.
        - ``'batched'`` : Useful for large numbers of light curves with
          relatively few shared filters, so ``'folded'`` loses large amounts of
          multi-epoch information. This mode splits the filters into sets of
          overlapping batches, runs each batch and uses the weighted means of
          each batch to produce a final result.
        Default is ``'auto'``.

    max_batch_parameters : int, optional
        The maximum number of parameters to use in a single retrieval when
        using ``'folded'`` or ``'batched'`` fitting modes. Default is 25.

    batch_overlap : in, optional
        The number of filters or epochs to overlap adjacent batches by where
        possible. This ensures that adjacent batches share information. Default
        is 2.

    host_T : tuple or None, optional
        The effective temperature of the host star, in Kelvin. Should be given
        as a (value, uncertainty) pair. Required if ld_fit_method is
        ``'single'`` or ``'coupled'``. Default is None.

    host_logg : tuple or None, optional
        The log_10 of the surface gravity of the host star, with gravity
        measured in cm/s2. Should be given as a (value, uncertainty) pair.
        Required if ld_fit_method is ``'single'`` or ``'coupled'``. Default is
        None

    host_z : tuple or None, optional
        The metalicity of the host, given as a (value, uncertainty) pair.
        Required if ld_fit_method is ``'single'`` or ``'coupled'``. Default is
        None

    host_r : tuple or None, optional
        The host radius in Solar radii, given as a (value, uncertainty) pair.
        Required for conversion of host-planet separation from AU to host radii

    nlive : int, optional
        The number of live points to use in the nested sampling retrieval.

    normalise : bool, optional
        If True, will assume that the light curves have not been normalised and
        will fit normalisation constants within the retrieval. The range to
        fit normalisation constants c_n are automatically detected using
            ``1/f_max <= c_n <= 1/f_min``
        as the default range, where ``f_min`` and ``f_max`` are the minimum and
        maximum flux values for a given light curve. Default is ``True``.

    detrend : bool, optional
        If False, no detrending will be attempted, even if specified by
        detrending list. Default is ``True``

    dlogz : float, optional
        Retrieval iteration will stop when the estimated contribution of the
        remaining prior volume to the total evidence falls below this
        threshold. Explicitly, the stopping criterion is
        ``ln(z + z_est) - ln(z) < dlogz``, where ``z`` is the current evidence
        from all saved samples and z_est is the estimated contribution from
        the remaining volume. The default is ``1e-3 * (nlive - 1) + 0.01``.

    maxiter : int or None, optional
        The maximum number of iterations to run. If None, will
        continue until stopping criterion is reached. Default is ``None``.

    maxcall : int or None, optional
        The maximum number of likelihood calls in retrieval. If None, will
        continue until stopping criterion is reached. Default is ``None``.

    dynesty_sample : str, optional
        Method used to sample uniformly within the likelihood constraint,
        conditioned on the provided bounds. Unique methods available are:
        uniform sampling within the bounds(``'unif'``), random walks with fixed
        proposals (``'rwalk'``), random walks with variable (“staggering”)
        proposals (``'rstagger'``), multivariate slice sampling along preferred
        orientations (``'slice'``), “random” slice sampling along all
        orientations (``'rslice'``), “Hamiltonian” slices along random
        trajectories (``'hslice'``), and any callable function which follows
        the pattern of the sample methods defined in dynesty.sampling.
        'auto' selects the sampling method based on the dimensionality of
        the problem (from ``ndim``). When ndim < 10, this defaults to ``'unif'``.
        When 10 <= ``ndim`` <= 20, this defaults to ``'rwalk'``. When ndim > 20,
        this defaults to ``'hslice'`` if a gradient is provided and ``'slice'``
        otherwise. ``'rstagger'`` and ``'rslice'`` are provided as alternatives for
        ``'rwalk'`` and ``'slice'``, respectively. Default is ``'rslice'``.

    dynesty_bounding : {``'none'``, ``'single'``, ``'multi'``, ``'balls'``, ``'cubes'``}, optional
        The decomposition to use in sampling. Default is 'multi'

    results_output_folder : str, optional
        Folder to save results to. TransitFit will create subfolders within
        this if folded or batched runs are being used. Default is
        ``'./output_parameters'``

    fitted_lightcurve_folder : str, optional
        The folder to save fitted light curves to. These files contain the
        normalised and detrended light curves, as well as the best fit curve.
        Default is ``'./fitted_lightcurves'``

    summary_file : str, optional
        The file to save the summarised final parameter results to. These are
        calculated by taking weighted averages over any batched fitting.
        Default is ``'summary_output.csv'``

    full_output_file : str, optional
        The file to save the full, non-summarised results to. This file gives
        the results for each batch, without averaging over batches to get
        summaried results. Default is ``'full_output.csv'``

    plot_folder : str, optional
        Path to folder to save plots to. Default is ``'./plots'``

    plot : bool, optional
        If ``True``, will plot all fitted light curves within the fitting routine,
        including any from partial fitting (eg, single filter modes). Default
        is ``True``.

    marker_color : matplotlib color, optional
        The colour to plot data points on plots. Default is ``'dimgray'``.

    line_colour : matplotlib color, optional
        The colour to plot best fit light curves on plots. Default is ``'black'``.

    ldtk_cache : str, optional
        This is the path to cache LDTK files to. If not specified, will
        default to the LDTK default.

    n_ld_samples : int, optional
        Controls the number of samples taken by PyLDTk when calculating LDCs
        when using ``'coupled'`` or ``'single'`` modes for limb darkening
        fitting. Default is ``20000``

    do_ld_mc : bool, optional
        If ``True``, will use MCMC sampling to more accurately estimate the
        uncertainty on intial limb darkening parameters provided by PyLDTk.
        Default is ``False``.

    data_skiprows : int, optional
        The number of rows to skip when reading in light curve data from a .txt
        file. Default is ``0``.

    allow_ttv : bool, optional
        If ``True``, will fit t0 for each epoch individually. Default is
        ``False``.

    filter_delimiter : str, optional
        The delimiter in filter profile files. Default is ``None``, which will
        lead to ``pandas`` trying to auto detect the delimiter.

    detrending_limits : list, optional
        The bounds on detrending coefficients, given as (lower, upper) pair for
        each detrending method. IF not provided, will default to ±10

    bin_data : bool, optional
        If True, any folded light curves will be plotted with data binned to an
        observing cadence given by `cadence`. Default is False.

    cadence : float, optional
        The observing cadence, in minutes, to bin data to if `bin_data` is
        True. Default is 2 (mirroring TESS observations)

    binned_color : str, optional
        The color to use for binned data. Default is `'red'`.

    Returns
    -------
    results : dict
        The results returned by ``Retriever.run_dynesty()``
    '''
    # Load in the data and work out number of telescopes, filters, and epochs
    lightcurves, detrending_index_array = read_input_file(data_files)

    n_telescopes = lightcurves.shape[0]
    n_filters = lightcurves.shape[1]
    n_epochs = lightcurves.shape[2]

    # Set up the Retriever
    retriever = Retriever(data_files, priors, n_telescopes, n_filters, n_epochs,
                          filter_info, detrending_list, limb_darkening_model,
                          host_T, host_logg, host_z, host_r, ldtk_cache,
                          data_skiprows, n_ld_samples, do_ld_mc, allow_ttv,
                          filter_delimiter, detrending_limits)

    # Run the retrieval!
    results = retriever.run_retrieval(ld_fit_method, fitting_mode,
                                      max_batch_parameters, maxiter, maxcall,
                                      dynesty_sample, nlive, dlogz, plot,
                                      results_output_folder,
                                      final_lightcurve_folder, summary_file,
                                      full_output_file, plot_folder,
                                      marker_color, line_color, dynesty_bounding, normalise,
                                      detrend, batch_overlap, bin_data, cadence, binned_color,
                                      walks, slices)

    return results



'''
Some detrending functions for use with LightCurve
'''

import numpy as np


class NthOrderDetrendingFunction:
    '''
    Arbitrary order detrending function which conserves flux

    Parameters
    ----------
    order : int
        The detrending function order. Must be greater than 0.
    '''
    def __init__(self, order):

        order = int(order)
        if not order > 0:
            raise ValueError('Order must be greater than 0')

        self.order = order


    def __call__(self, times, *args):
        '''
        Returns detrending values for the times
        '''
        if not len(args) == self.order:
            raise ValueError('Number of arguments {} supplied does not match the order {} of the detrending function.'.format(len(args), self.order))

        vals = np.zeros(len(times))
        for i in range(0, self.order):
            vals += args[i] * (times ** (i+1) - np.mean(times ** (i+1)))

        return vals

linear = NthOrderDetrendingFunction(1)
quadratic = NthOrderDetrendingFunction(2)



'''
Class to handle limb darkening

'''
import numpy as np
from ._ldtk_handler import LDTKHandler

class LimbDarkeningHandler:
    '''
    The LimbDarkeningHandler is designed to convert fitting parameters in
    the range [0,1] and convert them into physically allowed values for
    limb darkening coefficients for different models. This conversion is
    based on Kipping 2013 https://arxiv.org/abs/1308.0009.

    Parameters
    ----------
    default_model : str
        The default limb darkening model to use. Accepted values are
            - 'linear'
            - 'quadratic'
            - 'squareroot'
            - 'power2'
            - 'nonlinear'
        This model will be the default for conversions unless otheriwse
        specified.
    low_lim : float, optional
        The lower limit to use in conversion in the case where there are
        open bounds on a coefficient (power2 and nonlinear models). Note
        that in order to conserve sampling density in all regions for the
        power2 model, you should set lower_lim=-high_lim. Default is -5
    high_lim : float, optional
        The upper limit to use in conversion in the case where there are
        open bounds on a coefficient (power2 and nonlinear models). Note
        that in order to conserve sampling density in all regions for the
        power2 model, you should set lower_lim=-high_lim. Default is 5
    '''
    
    def __init__(self, default_model, low_lim=-5, high_lim=5):
        if default_model not in ['linear','quadratic','squareroot','power2','nonlinear']:
            raise ValueError('Unrecognised limb darkening model {}'.format(default_model))

        self.default_model = default_model

        # the default high and low bounds to use for unconstrained values
        self.high = high_lim
        self.low = low_lim

        self.ldtk_handler = None

        self.host_T = None
        self.host_logg = None
        self.host_z = None

    def get_required_coefficients(self, model=None):
        '''
        Finds the number of coefficients required for a given model and returns
        a string name for each of the form 'uX' where X is a number.

        Parameters
        ----------
        model : str or None, optional
            Model to get the coefficients for. Default is self.default_model

        Returns
        -------
        coeficient names : list of str
            The coefficient names
        '''
        if model is None:
            model = self.default_model

        if model == 'linear':
            return ['q0']
        if model in ['quadratic','squareroot', 'power2']:
            return ['q0', 'q1']
        if model == 'nonlinear':
            return ['q0', 'q1', 'q2', 'q3']

        raise ValueError('Unrecognised model {}'.format(model))

    def convert_qtou(self, *q, model=None):
        '''
        Takes parameters q distributed between [0,1] and converts them into
        physical values for limb darkening coefficients.

        Conversions for quadratic, square root, and logarithmic are from
        Kipping 2013 https://arxiv.org/abs/1308.0009 for two-parameter limb
        darkening methods

        Notes
        -----
        This is the inverse of convert_utoq
        '''
        if model is None:
            model = self.default_model

        if model == 'linear':
            # coefficient is limited to 0<A<1 by Kipping criteria
            return q[0]

        if model == 'quadratic':
            return 2 * np.sqrt(q[0]) * q[1], np.sqrt(q[0]) * (1 - 2 * q[1])

        if model == 'squareroot':
            return np.sqrt(q[0]) * (1 - 2 * q[1]), 2 * np.sqrt(q[0]) * q[1]

        if model == 'power2':
            # This parameterisation has been derived for TransitFit and is
            # explained in the paper
            u1 = q[0] * (1 - self.low) + self.low

            if u1 < 0:
                # negative quadrant
                return u1, self.low * (1 - q[1])

            else:
                # positive quadrant
                return u1, self.high * q[1]

        if model == 'nonlinear':
            # This is an 'outstanding and formidable problem' to apply the
            # Kipping (2013) method to (see Kipping 2016) and as such, we can
            # only fit for free parameters, without reparameterising to allow
            # only physically valid combinations.
            u1 = q[0] * (self.high - self.low) + self.low
            u2 = q[1] * (self.high - self.low) + self.low
            u3 = q[2] * (self.high - self.low) + self.low
            u4 = q[3] * (self.high - self.low) + self.low

            return u1, u2, u3, u4

        raise ValueError('Unrecognised model {}'.format(model))

    def convert_utoq(self, *u, model=None):
        '''
        Takes actual values of the LD coefficients and converts them to a
        value q between [0,1].

        Conversions for quadratic, square root, and logarithmic are from
        Kipping 2013 https://arxiv.org/abs/1308.0009 for two-parameter limb
        darkening methods

        Notes
        -----
        This is the inverse of convert_qtou
        '''
        if model is None:
            model = self.default_model

        if model == 'linear':
            # coefficient is limited to 0<A<1 by Kipping criteria
            return q[0]

        if model == 'quadratic':
            return (u[0] + u[1]) ** 2, u[0]/(2 * (u[0] + u[1]))

        if model == 'squareroot':
            return (u[0] + u[1]) ** 2, u[1]/(2 * (u[0] + u[1]))

        if model == 'power2':
            # This parameterisation has been derived for TransitFit and is
            # explained in the paper
            q1 = (u[0] - self.low)/(1 - self.low)

            if u[0] < 0:
                # negative quadrant
                return u1, 1 - (u[1]/self.low)

            else:
                # positive quadrant
                return u1, u[1]/self.high

        if model == 'nonlinear':
            # This is an 'outstanding and formidable problem' to apply the
            # Kipping (2013) method to (see Kipping 2016) and as such, we can
            # only fit for free parameters, without reparameterising to allow
            # only physically valid combinations.
            q0 = (u[0] - self.low) /  (self.high - self.low)
            q1 = (u[1] - self.low) /  (self.high - self.low)
            q2 = (u[2] - self.low) /  (self.high - self.low)
            q3 = (u[3] - self.low) /  (self.high - self.low)

            return q0, q1, q2, q3

        raise ValueError('Unrecognised model {}'.format(model))

    def convert_qtou_with_errors(self, q, q_err, model=None):
        '''
        Converts kipping parameters with errors into physical LDCs with errors
        '''
        if model is None:
            model = self.default_model

        u = list(self.convert_qtou(*q, ))

        if model == 'linear':
            u_err = q_err

            return u, u_err

        if model  == 'quadratic':
            u_err = []
            u_err.append(np.sqrt(((q[1] * q_err[0])**2)/q[0] + 4 * q[0] * q_err[1]**2))
            u_err.append(np.sqrt((((1- 2 * q[1]) * q_err[0])**2)/(0.25 * q[0]) + 4 * q[0] * q_err[1] ** 2))

            return u, u_err

        if model == 'squareroot':
            u_err = []
            u_err.append(np.sqrt((((1- 2 * q[1]) * q_err[0])**2)/(0.25 * q[0]) + 4 * q[0] * q_err[1] ** 2))
            u_err.append(np.sqrt(((q[1] * q_err[0])**2)/q[0] + 4 * q[0] * q_err[1]**2))

            return u, u_err

        if model == 'power2':
            u_err = []
            u_err.append(abs(1-self.low) * q_err[0])
            if u[1] < 0:
                u_err.append(abs(self.low) * q_err[1])
            else:
                u_err.append(abs(self.high) * q_err[1])

            return u, u_err

        if model == 'nonlinear':
            u_err = [abs(self.high - self.low) * err for err in q_err]

            return u, u_err


    def initialise_ldtk(self, host_T, host_logg, host_z, filters,
                        model=None, n_samples=20000, do_mc=False,
                        cache_path=None):
        '''
        Sets up an LDTKHandler to deal with interfacing between ldtk and
        TransitFit

        Parameters
        ----------
        host_T : tuple
            The effective temperature of the host star, in Kelvin, given as a
            (value, uncertainty) pair.
        host_logg : tuple
            The log_10 of the surface gravity of the host star, with gravity
            measured in cm/s2. Should be given as a (value, uncertainty) pair.
        host_z : tuple
            The metalicity of the host, given as a (value, uncertainty) pair.
        filters : array_like
            The set of filters, given in [low, high] limits for the wavelengths
            with the wavelengths given in nanometers. The ordering of the
            filters should correspond to the filter_idx parameter used
            elsewhere.
        ld_method : str or None, optional
            The model of limb darkening to use. Allowed values are 'linear',
            'quadratic', 'squareroot', 'power2', and 'nonlinear'. If None, will
            default to self.default_model. Default is None.
        n_samples : int, optional
            The number of limb darkening profiles to create. Passed to
            ldtk.LDPSetCreator.create_profiles(). Default is 20000.
        do_mc : bool, optional
            If True, will use MCMC to estimate coefficient uncertainties more
            accurately. Default is False.
        cache_path : str, optional
            This is the path to cache LDTK files to. If not specified, will
            default to the LDTK default
        '''
        if model is None:
            model = self.default_model

        self.host_T = host_T
        self.host_logg = host_logg
        self.host_z = host_z

        self.ldtk_handler = LDTKHandler(host_T, host_logg, host_z, filters,
                                        model, n_samples, do_mc, cache_path)


    def ldtk_lnlike(self, coeffs, model=None):
        '''
        Uses LDTK to evaluate the log likelihood for a set of coefficients.
        Note that coeffs should be the values returned by convert_qtoA
        and NOT the unit values.

        Parameters
        ----------
        coeffs : array_like, shape (n_filters, n_coeffs)
            The coefficients to evaluate the log likelihood for.
        ld_model : str, optional
            The model to use. Defaults to self.default_model

        Returns
        -------
        lnlike : float
            The log likelihood of the coefficients
        '''

        if self.ldtk_handler is None:
            raise Exception('ldtk_handler is not initialised. Please initialise using initialise_ldtk()')

        if model is None:
            model = self.default_model

        return self.ldtk_handler.lnlike(coeffs, model)

    def ldtk_estimate(self, ld0_values, model=None):
        '''
        Uses LDTK to estimate the LDCs for all filters when given a set of LDCs
        for filter 0, based on the ratios between the best values
        found in initialisation.

        Parameters
        ----------
        ld1_values : float or array_like
            The LD parameters for filter 0
        ld_model : str, optional
            The limb darkening model to use. Defaults to self.default_model

        Returns
        -------
        all_ld_values : array_like, shape (n_filters, n_coeffs)
            The estimated limb darkening parameters
        '''
        if self.ldtk_handler is None:
            raise Exception('ldtk_handler is not initialised. Please initialise using initialise_ldtk()')

        if model is None:
            model = self.default_model

        return self.ldtk_handler.estimate_values(ld0_values, model)
